import pytest
import os
from ave.display.curses_screen import CursesScreen

winskip = pytest.mark.skipif(
    os.name == "nt", reason="CircleCI cannot run this test on windows")


@winskip
def make_screen():
    try:
        return CursesScreen()
    except:  # noqa: E722
        return CursesScreen(False)


@winskip
def close_screen(s):
    try:
        s.close()
    except:  # noqa: E722
        s.close(False)


@winskip
def test_start_and_close():
    s = make_screen()
    close_screen(s)


@winskip
def test_clear():
    s = make_screen()
    s.clear()
    close_screen(s)


@winskip
@pytest.mark.parametrize('file', ["credits", "title", "user"])
def test_print_file(file):
    s = make_screen()
    s.print_file(file)
    s.close(False)


@winskip
@pytest.mark.parametrize('inventory', [
    [], ["hat", "shoes"], ["a" * 100], ["a"] * 100])
def test_show_inventory(inventory):
    s = make_screen()
    s.show_inventory([])
    s.show_inventory(["hat", "shoes"])
    s.show_inventory(["hat"] * 100)
    close_screen(s)



import pytest
import re
import ave
from ave.parsing import string_functions as sf


@pytest.mark.parametrize('txt', [
    "",
    "This is an option => gohere ? hat",
    "This option has an escape <| => |> gohere ?! hat",
    "A<|???|> B C <|D=>E|>"
])
def test_unescape_escape_string(txt):
    assert sf.unescape(sf.escape(txt)) == re.sub(r"<\|(.*?)\|>", r"\1", txt)


def test_escape_string():
    assert "=>" not in sf.escape("test <|=>|> escaping")
    assert "=>" in sf.escape("test => escaping")
    assert "=>" in sf.escape("test <|a|> => <|???|>escaping")


def test_between():
    assert sf.between("abcde(fgh)", "(", ")") == "fgh"
    assert sf.between("abcde(fgh)ijk(lmn)", "(", ")") == "fgh"


def test_replacements():
    assert sf._replacements("ab %v%") == "ab " + ave.__version__


def test_more_unescape():
    assert sf.unescape(sf.escape("$var$")) == "$var$"
    assert sf.unescape(sf.escape("<|$var$|>")) != "$var$"
    assert sf.more_unescape(sf.escape("<|$var$|>")) == "$var$"



from ave import AVE
from ave import load_game_from_library


def test_game_library():
    ave = AVE()
    ave.get_download_menu()


def test_load_game_from_library():
    ave = AVE()
    game = load_game_from_library(ave.get_download_menu()[0][2])
    game.load()
    assert game["start"].id != "fail"



"""Number classes."""

import random


class Number:
    """Base number class."""

    def get_value(self, character):
        """Get the value of the Number."""
        raise NotImplementedError()

    def get_all_variables(self):
        """Get a list of all variables used in this expression."""
        raise NotImplementedError()


class Constant(Number):
    """A constant."""

    def __init__(self, value):
        """Make the number."""
        self.value = value

    def get_value(self, character):
        """Get the value of the Number."""
        return self.value

    def get_all_variables(self):
        """Get a list of all variables used in this expression."""
        return []


class Sum(Number):
    """The sum of multiple Numbers."""

    def __init__(self, *items):
        """Make the number."""
        self.items = items

    def get_value(self, character):
        """Get the value of the Number."""
        return sum(i.get_value(character) for i in self.items)

    def get_all_variables(self):
        """Get a list of all variables used in this expression."""
        out = []
        for i in self.items:
            out += i.get_all_variables()
        return out


class Product(Number):
    """The product of multiple Numbers."""

    def __init__(self, *items):
        """Make the number."""
        self.items = items

    def get_value(self, character):
        """Get the value of the Number."""
        out = 1
        for i in self.items:
            out *= i.get_value(character)
        return out

    def get_all_variables(self):
        """Get a list of all variables used in this expression."""
        out = []
        for i in self.items:
            out += i.get_all_variables()
        return out


class Division(Number):
    """The result of a division."""

    def __init__(self, value, *items):
        """Make the number."""
        self.value = value
        self.items = items

    def get_value(self, character):
        """Get the value of the Number."""
        out = self.value.get_value(character)
        for i in self.items:
            out /= i.get_value(character)
        return out

    def get_all_variables(self):
        """Get a list of all variables used in this expression."""
        out = []
        for i in self.items:
            out += i.get_all_variables()
        return out


class Negative(Number):
    """The negative of another Number."""

    def __init__(self, item):
        """Make the number."""
        self.item = item

    def get_value(self, character):
        """Get the value of the Number."""
        return -self.item.get_value(character)

    def get_all_variables(self):
        """Get a list of all variables used in this expression."""
        return self.item.get_all_variables()


class Variable(Number):
    """The value of a variable."""

    def __init__(self, item):
        """Make the number."""
        self.item = item

    def get_value(self, character):
        """Get the value of the Number."""
        return character.numbers[self.item]

    def get_all_variables(self):
        """Get a list of all variables used in this expression."""
        return [self.item]


class Random(Number):
    """A random number."""

    def __init__(self, *args):
        """Make the number."""
        if len(args) == 0:
            self.start = Constant(0)
            self.end = Constant(1)
        elif len(args) == 1:
            self.start = Constant(0)
            self.end = args[0]
        elif len(args) == 2:
            self.start = args[0]
            self.end = args[1]

    def get_value(self, character):
        """Get the value of the Number."""
        start = self.start.get_value(character)
        end = self.end.get_value(character)
        size = end - start
        return start + random.random() * size

    def get_all_variables(self):
        """Get a list of all variables used in this expression."""
        return self.start.get_all_variables() + self.end.get_all_variables()



"""Functions to manipulate strings."""

import re
from .. import config

symbols = {
    "!?": "SYMBOLIINNTTEERROOBBAANNGG",
    "=>": "SYMBOLAARROOWW",
    "?": "SYMBOLQQUUEESSTTIIOONN",
    "+": "SYMBOLPPLLUUSS",
    "~": "SYMBOLTTIILLDDEE",
    "=": "SYMBOLEEQQUUAALL",
    "-": "SYMBOLDDAASSHH",
    "*": "SYMBOLAASSTTEERRIISSKK",
    "/": "SYMBOLFFWWDDSSLLAASSHH",
    "@": "SYMBOLAATT",
    "%": "SYMBOLPPEERRCCEENNTT",
    "#": "SYMBOLHHAASSHH",
    "_": "SYMBOLUUNNDDEERRSCCOORREE",
    "[": "SYMBOLOOPPEENNSQ",
    "]": "SYMBOLCCLLOOSSEESQ",
    "(": "SYMBOLOOPPEENNRO",
    ")": "SYMBOLCCLLOOSSEERO"
}

more_symbols = {
    "$": "SYMBOLDDOOLLAARR",
    "<": "SYMBOLOOPPEENNPO",
    ">": "SYMBOLCCLLOOSSEEPO"
}

expression_symbols = {
    "+": "EXPRESSIONPPLLUUSS",
    "-": "EXPRESSIONMMIINNUUSS",
    "*": "EXPRESSIONTTIIMMEESS",
    "/": "EXPRESSIONDDIIVVIIDDEE",
    ",": "EXPRESSIONCCOOMMMMAA"
}


def between(text, pre, post):
    """Get the part of a string between two substrings.

    Parameters
    ----------
    test : string
        The string to search
    pre : string
        The substring before the selection.
    post : string
        The substring after the selection.

    Returns
    -------
    string
        The text between the substrings, or an empty string if not found
    """
    if pre not in text or post not in text:
        return ""
    else:
        return text.split(pre, 1)[1].split(post, 1)[0]


def _replacements(txt):
    txt = config.version.join(txt.split("%v%"))
    return txt


def clean(txt):
    """Strip spaces and newlines and insert data."""
    txt = txt.strip()
    txt = txt.replace("<newline>", "\n")
    return _replacements(txt)


def clean_newlines(txt):
    """Strip newlines and insert data such as the AVE version number."""
    txt = txt.strip("\n")
    txt = txt.replace("<newline>", "\n")
    return _replacements(txt)


def _escape(matches):
    text = matches[1]
    for i, j in symbols.items():
        text = text.replace(i, j)
    for i, j in more_symbols.items():
        text = text.replace(i, j)
    return text


def escape(txt):
    """Escape text between <| and |>."""
    return re.sub(r"<\|(.*?)\|>", _escape, txt)


def unescape(text):
    """Restore text between <| and |> that was escaped."""
    for i, j in symbols.items():
        text = text.replace(j, i)
    return text


def more_unescape(text):
    """Restore final bits of text between <| and |> that was escaped.

    These are done later as they would otherwise be changed when getting
    room text. For example <| $variable$ |> would be replaces with the
    variable value if it was unescaped earlier.
    """
    for i, j in more_symbols.items():
        text = text.replace(j, i)
    return text


def finalise(txt, numbers, currency=None):
    """Insert variables into text, then unescape final characters."""
    for i, n in numbers.items():
        txt = txt.replace("$" + i + "$", str(n))
    if currency is not None:
        txt = txt.replace("<money>", currency)
    return more_unescape(txt)


def _escape_expr(matches):
    text = matches[1]
    for i, j in expression_symbols.items():
        text = text.replace(i, j)
    return "{|" + text + "|}"


def escape_expression(txt):
    """Escape text between ( and )."""
    pre = ""
    while pre != txt:
        pre = txt
        txt = re.sub(r"\(([^\(\)\|](?:[^\(\)]*[^\(\)\|])?)\)",
                     _escape_expr, txt)
    return txt


def unescape_expression(text):
    """Restore text between ( and ) that was escaped."""
    for i, j in expression_symbols.items():
        text = text.replace(j, i)
    text = text.replace("{|", "(")
    text = text.replace("|}", ")")
    return text



"""Exceptions used by AVE."""


class AVEQuit(BaseException):
    """Quit AVE."""


class AVEAgain(BaseException):
    """Play the same game again."""


class AVEToMenu(BaseException):
    """Leave to the main menu."""


class AVEGameOver(BaseException):
    """Game over."""


class AVEWinner(BaseException):
    """You win."""


class AVEVersionError(BaseException):
    """Your version of AVE cannot play this game."""


class AVENoInternet(BaseException):
    """AVE could not load an online resource."""



"""Functions to run AVE."""

import os
import json
from ave import AVE, Character, config


def run():
    """Run AVE in terminal."""
    import argparse

    parser = argparse.ArgumentParser(description='Run the AVE game engine.')
    parser.add_argument('folder', metavar='folder', nargs='?', default=None,
                        help='A folder to load AVE games from.')

    args = parser.parse_args()

    from .display.curses_screen import CursesScreen
    ave = AVE(screen=CursesScreen(), character=Character())
    ave.load_games_from_json(os.path.join(config.ave_folder, "gamelist.json"))
    if args.folder is not None:
        ave.load_games(args.folder, "[*] ")
    ave.start()


def make_json(game_folder=config.games_folder, json_folder=config.ave_folder):
    """Make a json containing metadata for every game and write it to file."""
    config.debug = True
    ave = AVE()
    ave.load_games(game_folder)
    gamelist = generate_json(game_folder)
    with open(os.path.join(json_folder, "gamelist.json"), "w") as f:
        json.dump(gamelist, f)


def generate_json(game_folder=config.games_folder):
    """Make a json containing metadata for every game."""
    config.debug = True
    ave = AVE()
    ave.load_games(game_folder)
    return [{
        "title": game.title,
        "author": game.author,
        "desc": game.description,
        "active": game.active,
        "version": game.version,
        "ave_version": game.ave_version,
        "filename": game.filename,
        "number": game.number
    } for game in ave.games]


def test_game():
    """Test an AVE game for errors."""
    import argparse

    parser = argparse.ArgumentParser(
        description='Test an AVE game for errors.')
    parser.add_argument('file', metavar='file',
                        help='A file to test')

    args = parser.parse_args()

    from ave import load_game_from_file
    from ave.test import check_game

    game = load_game_from_file(args.file)
    game.load()

    errors = check_game(game)

    for i in range(5, 0, -1):
        ls = [e for e in errors if e.error_value == i]
        if len(ls) > 0:
            print("")
            for e in ls:
                print(e)



"""The AVE and GameLibrary classes that run AVE in a terminal."""

import os
import json

from . import config
from .exceptions import (AVEGameOver, AVEWinner, AVEToMenu, AVEQuit,
                         AVENoInternet)
from .game import Game
from .parsing.game_loader import (
    load_game_from_file, load_library_json, load_game_from_library)


class AVE:
    """The AVE class that runs the Character, Screen and Game."""

    def __init__(self, screen=None, character=None):
        """Create an AVE class.

        Parameters
        ----------
        screen : ave.display.base.Screen
            The screen that will display the game
        character : ave.game.Character
            The character playing the game
        """
        self.screen = screen
        self.character = character
        self.games = None
        self.items = None
        self._unsorted_games = []

    def start(self):
        """Start running the main AVE loop."""
        if config.debug:
            self._debug_start()
        else:
            self._actual_start()

    def _debug_start(self):
        """Start AVE in debug mode.

        In debug mode, exceptions will be diaplayed on exit.
        """
        try:
            self.show_title_screen()
        except AVEQuit:
            self.exit()
            print("Goodbye...")
        finally:
            self.exit()

    def _actual_start(self):
        """Start AVE in standard mode."""
        while True:
            try:
                self.show_title_screen()
            except AVEQuit:
                self.exit()
                print("Goodbye...")
                break
            except:  # noqa: E722
                self.exit()
                break

    def show_title_screen(self):
        """Display the AVE title screen."""
        self.screen.print_titles()
        game_to_load = self.screen.title_menu(
            self.games.titles() + ["- user contributed games -"])
        if game_to_load == len(self.games.titles()):
            self.show_download_menu()
        else:
            the_game = self.games[game_to_load]
            self.run_the_game(the_game)

    def sort_games(self):
        """Remove disabled games and sort the games by number."""
        ordered_games = {}
        other_games = []
        for g in self._unsorted_games:
            if config.version_tuple < g.ave_version:
                continue
            if g.active or config.debug:
                if g.number is None:
                    other_games.append(g)
                else:
                    assert g.number not in ordered_games
                    ordered_games[g.number] = g
        other_games.sort(key=lambda x: x.id)
        self.games = GameLibrary([
            ordered_games[i]
            for i in sorted(ordered_games.keys())] + other_games)

    def load_games(self, folder, prefix=""):
        """Load the metadata of games from a folder.

        Parameters
        ----------
        folder: str
            The folder
        """
        for game in os.listdir(folder):
            if game[-4:] == ".ave":
                g = load_game_from_file(os.path.join(folder, game),
                                        game)
                g.title = prefix + g.title
                self.add_game(g)
        self.sort_games()

    def add_game(self, game):
        """Add a game to the (unsorted) game list."""
        if game.file is not None:
            for g in self._unsorted_games:
                if g.file == game.file:
                    return
        if game.url is not None:
            for g in self._unsorted_games:
                if g.url == game.url:
                    return
        self._unsorted_games.append(game)

    def load_games_from_json(self, json_file):
        """Load the metadata of games from a json.

        Parameters
        ----------
        json_file: str
            The location of the json file
        """
        with open(json_file) as f:
            gamelist = json.load(f)
        for game in gamelist:
            self.add_game(Game(
                file=os.path.join(config.games_folder, game["filename"]),
                title=game["title"], number=game["number"],
                description=game["desc"],
                author=game["author"], active=game["active"],
                version=game["version"], filename=game["filename"],
                ave_version=tuple(game["ave_version"])))
        self.sort_games()

    def get_download_menu(self):
        """Get the list of games from the online library.

        Returns
        -------
        list
            A list of the title, author and local url for each game.
        """
        try:
            library = load_library_json()
        except AVENoInternet:
            self.no_internet()
            raise AVEToMenu
        return [(game["title"], game["author"], i)
                for i, game in enumerate(library)]

    def show_download_menu(self):
        """Show a menu of games from the online library."""
        try:
            self.screen.print_download()
            menu_items = self.get_download_menu()
            game_n = self.screen.download_menu(
                [a[0] + ' by ' + a[1] for a in menu_items])
            the_game = load_game_from_library(menu_items[game_n][2])
            self.run_the_game(the_game)
        except AVEToMenu:
            self.show_title_screen()

    def no_internet(self):
        """Show a "you have no internet" notification."""
        self.screen.no_internet()
        self.screen.wait_for_input()

    def run_the_game(self, the_game):
        """Run a game.

        Parameters
        ----------
        the_game : ave.game.Game
            The game
        """
        the_game.load()
        again = True
        while again:
            again = False
            try:
                self.character.reset(the_game.items)
                self.screen.show_titles(
                    the_game.title,
                    the_game.description,
                    the_game.author,
                    the_game.version)
                self.game_loop(the_game)
            except AVEGameOver:
                next = self.screen.gameover()
                if next == 0:
                    again = True
                if next == 2:
                    raise AVEQuit
            except AVEWinner:
                next = self.screen.winner()
                if next == 0:
                    again = True
                if next == 2:
                    raise AVEQuit
            except AVEToMenu:
                pass

    def game_loop(self, the_game):
        """Run the game in a loop so it can be re-played.

        Parameters
        ----------
        the_game : ave.game.Game
            The game
        """
        while True:
            text, options = the_game.get_room_info(
                self.character, the_game.currency)
            self.screen.clear()
            self.screen.put_ave_logo()
            self.screen.show_inventory(
                self.character.get_inventory(the_game.items,
                                             the_game.currency))
            self.screen.type_room_text(text)
            next_id = self.screen.menu(list(options.values()))
            the_game.pick_option(
                list(options.keys())[next_id],
                self.character)

    def exit(self):
        """Exit AVE."""
        self.screen.close()


class GameLibrary:
    """The GameLibrary class that stores the list of available games."""

    def __init__(self, games):
        """Make a GameLibrary.

        Parameters
        ----------
        games : list
            A list of games.
        """
        self.games = games

    def titles(self):
        """Get the title of each game.

        Returns
        -------
        list
            The title of each game
        """
        return [g.title for g in self.games]

    def descriptions(self):
        """Get the description of each game.

        Returns
        -------
        list
            The description of each game
        """
        return [g.description for g in self.games]

    def titles_and_descriptions(self):
        """Get the title and description of each game.

        Returns
        -------
        iterator
            The title and description of each game
        """
        return zip(self.titles(), self.descriptions())

    def game(self, n):
        """Get the nth game.

        Parameters
        ----------
        n : int
            n
        """
        return self.games[n]

    def __getitem__(self, n):
        """Get the nth game.

        Parameters
        ----------
        n : int
            n
        """
        return self.games[n]



"""Error handlers."""


class AVEErrorHandler:
    """Class to store errors found when checking games."""

    def __init__(self, description, error_type, error_value):
        """Make the error handler."""
        self.description = description
        self.error_type = error_type
        self.error_value = error_value

    def __str__(self):
        """Return a string describing the error."""
        return self.error_type + ": " + self.description


class AVEFatalError(AVEErrorHandler):
    """This error is so bad that the game will not run."""

    def __init__(self, description):
        """Make the error handler."""
        super().__init__(description, "Fatal Error", 5)


class AVEError(AVEErrorHandler):
    """There is an error but part of the game will still run."""

    def __init__(self, description):
        """Make the error handler."""
        super().__init__(description, "Error", 4)


class AVEWarning(AVEErrorHandler):
    """The game will run, but there may is a small problem."""

    def __init__(self, description):
        """Make the error handler."""
        super().__init__(description, "Warning", 3)


class AVEInfo(AVEErrorHandler):
    """The game will run, but there maybe is a small problem."""

    def __init__(self, description):
        """Make the error handler."""
        super().__init__(description, "Info", 2)


class AVENote(AVEErrorHandler):
    """The game will run, but there maybe is a tiny problem."""

    def __init__(self, description):
        """Make the error handler."""
        super().__init__(description, "Note", 1)



"""Use curses to display the game in terminal."""

from ..exceptions import AVEToMenu, AVEQuit
from .. import config
from ..parsing.string_functions import clean_newlines
from .base import Screen
import curses

HEIGHT = 25
WIDTH = 80


class CursesScreen(Screen):
    """The Screen class that looks after curses."""

    def __init__(self, cbreak=True):
        """Make the screen."""
        self.stdscr = curses.initscr()

        curses.noecho()
        if cbreak:
            curses.cbreak()
        curses.curs_set(0)
        self.stdscr.keypad(1)

        try:
            # Resize terminal iff it is too small
            y, x = self.stdscr.getmaxyx()
            if y < HEIGHT or x < WIDTH:
                print("\x1b[8;" + str(HEIGHT + 2) + ";" + str(WIDTH + 2) + "t")
                curses.resizeterm(HEIGHT + 2, WIDTH + 2)
                self.stdscr.clear()
                self.stdscr.refresh()
        except AttributeError:
            # Windows
            pass

        curses.start_color()
        curses.use_default_colors()
        # curses.assume_default_colors(-1, curses.COLOR_MAGENTA)

        bg_color = curses.COLOR_BLACK
        if curses.can_change_color():
            curses.init_color(curses.COLOR_MAGENTA, 190, 40, 140)
            bg_color = curses.COLOR_MAGENTA

        # @ in title and credits
        curses.init_pair(1, curses.COLOR_BLACK, curses.COLOR_RED)
        # ^ in title and credits
        curses.init_pair(2, curses.COLOR_BLACK, curses.COLOR_GREEN)
        # = in title and credits
        curses.init_pair(3, curses.COLOR_BLACK, curses.COLOR_BLUE)
        # menu unselected
        curses.init_pair(4, curses.COLOR_BLACK, curses.COLOR_YELLOW)
        # menu selected
        curses.init_pair(5, curses.COLOR_WHITE, curses.COLOR_RED)
        # A in title
        curses.init_pair(6, curses.COLOR_RED, bg_color)
        # V in title
        curses.init_pair(7, curses.COLOR_GREEN, bg_color)
        # E in title
        curses.init_pair(8, curses.COLOR_BLUE, bg_color)
        # inventory
        curses.init_pair(9, curses.COLOR_WHITE, curses.COLOR_BLUE)
        # gameover
        curses.init_pair(10, curses.COLOR_WHITE, curses.COLOR_BLUE)
        # blank
        curses.init_pair(11, curses.COLOR_WHITE, bg_color)

        self.stdscr.refresh()

    def no_internet(self):
        """Display a "You have no internet" message."""
        stuff = []
        for i, c in enumerate("Unable to load from the internet. "
                              "Press <q> to go back."):
            stuff.append((0, i, c, curses.color_pair(8)))
        self.show(stuff, 15, 5, 1, 56)

    def press_enter(self):
        """Display a "Press <enter> to continue" message."""
        stuff = []
        for i, c in enumerate("Press <enter> to continue."):
            stuff.append((0, i, c, curses.color_pair(8)))
        self.show(stuff, 15, 5, 1, 56)

    def clear(self):
        """Clear the screen."""
        pad = self.newpad()
        pad.refresh(0, 0, 1, 1, HEIGHT, WIDTH)

    def close(self, cbreak=True):
        """Close the curses screen."""
        if cbreak:
            curses.nocbreak()
        curses.curs_set(1)
        self.stdscr.keypad(0)
        curses.echo()
        if cbreak:
            curses.endwin()

    def newpad(self, y=HEIGHT, x=WIDTH):
        """Make a new pad to write to the screen."""
        pad = curses.newpad(y + 1, x)
        for ypos in range(y):
            pad.addstr(ypos, 0, " " * x, curses.color_pair(11))
        return pad

    def put_ave_logo(self):
        """Print the AVE logo at the top of the screen."""
        stuff = [(0, 0, "A", curses.color_pair(6)),
                 (0, 1, "V", curses.color_pair(7)),
                 (0, 2, "E", curses.color_pair(8))]
        self.show(stuff, 0, WIDTH - 3, 2, 3)

    def print_file(self, filename):
        """Print the contents of a text file."""
        import os
        self.clear()
        stuff = []
        with open(os.path.join(config.screens_folder, filename)) as f:
            y = 0
            y_beg = None
            for line in f.readlines():
                if line[0] != "#":
                    line = clean_newlines(line)
                    for x, c in enumerate(line):
                        if x >= WIDTH:
                            break
                        if c == "@":
                            stuff.append((y, x, " ", curses.color_pair(1)))
                        elif c == "^":
                            stuff.append((y, x, " ", curses.color_pair(2)))
                        elif c == "=":
                            stuff.append((y, x, " ", curses.color_pair(3)))
                        elif c == "A" and len(line) > x + 1 \
                                and line[x + 1] == "V" \
                                and len(line) > x + 2 and line[x + 2] == "E":
                            stuff.append((y, x, "A", curses.color_pair(6)))
                        elif x >= 1 and line[x - 1] == "A" and c == "V" \
                                and len(line) > x + 1 and line[x + 1] == "E":
                            stuff.append((y, x, "V", curses.color_pair(7)))
                        elif x >= 2 and line[x - 2] == "A" \
                                and line[x - 1] == "V" and c == "E":
                            stuff.append((y, x, "E", curses.color_pair(8)))
                        else:
                            stuff.append((y, x, c, curses.color_pair(11)))
                    y += 1
                    if y >= HEIGHT:
                        y -= 1
                        break
                elif "# type" in line:
                    self.show(stuff, y=y + 1, x=WIDTH)
                    y_beg = y
                    stuff = []
                    y = 0
        if y_beg is None:
            self.show(stuff, y=y + 1, x=WIDTH)
        else:
            self.type(stuff, py=y_beg, y=y, x=WIDTH, title=True)

    def gameover(self):
        """Show the game over screen."""
        return self.gameend("GAME OVER")

    def winner(self):
        """Show the "you win" screen."""
        return self.gameend("YOU WIN!")

    def gameend(self, text):
        """Allow the user to play again or go back to the menu."""
        pad = self.newpad(10, WIDTH - 19)
        gst = " " * ((WIDTH - 29) // 2) + text
        gst += " " * (WIDTH - len(gst))
        gst = gst[:WIDTH - 20]
        pad.addstr(1, 0, " " * (WIDTH - 20), curses.color_pair(10))
        pad.addstr(2, 0, gst, curses.color_pair(10))
        pad.addstr(3, 0, " " * (WIDTH - 20), curses.color_pair(10))
        pad.addstr(4, 0, " " * (WIDTH - 20), curses.color_pair(10))
        pad.addstr(5, 0, " " * (WIDTH - 20), curses.color_pair(10))
        pad.addstr(6, 0, " " * (WIDTH - 20), curses.color_pair(10))
        pad.addstr(7, 0, " " * (WIDTH - 20), curses.color_pair(10))
        pad.refresh(0, 0, 4, 11, 12, WIDTH - 9)
        return self.gameover_menu(["Play again", "Play another game", "Quit"])

    def show_inventory(self, inventory):
        """Display the inventory."""
        pad = self.newpad(14, 20)
        pad.addstr(0, 0, "INVENTORY" + " " * 10, curses.color_pair(9))
        for i in range(12):
            if i < len(inventory):
                item = inventory[i]
                pad.addstr(
                    i + 1, 0,
                    "  " + item[:17] + " " * (17 - len(item)),
                    curses.color_pair(9))
            else:
                pad.addstr(i + 1, 0, " " * 19, curses.color_pair(9))
        pad.refresh(0, 0, 2, WIDTH - 19, 14, WIDTH)

    def type_room_text(self, text):
        """Type the room text."""
        y = 0
        x = 0
        stuff = []
        text = text.replace("\n", " \n ")
        for wn, word in enumerate(text.split(" ")):
            if word == "\n":
                y += 1
                x = 0
            elif word != "":
                if x + len(word) > WIDTH - 22:
                    y += 1
                    x = 0
                    if y > 13:
                        self.type(stuff)
                        self.press_enter()
                        self.wait_for_input(["ENTER", "q"])
                        return self.type_room_text(
                            " ".join(text.split(" ")[wn + 1:]))
                for i, c in enumerate(word):
                    stuff.append((y, x, c))
                    x += 1
                stuff.append((y, x, " "))
                x += 1
        self.type(stuff)

    def type(self, stuff, py=0, px=0, y=HEIGHT, x=WIDTH - 21,
             title=False):
        """Type some text."""
        from time import sleep
        pad = self.newpad(y, x)
        delay = True
        self.stdscr.nodelay(1)
        for char in stuff:
            if char[0] < y and char[1] < x:
                if self.stdscr.getch() != -1:
                    delay = False
                if char[2] != " " and (delay or title):
                    sleep(.01)
                if len(char) == 3:
                    pad.addch(char[0], char[1], char[2], curses.color_pair(11))
                if len(char) == 4:
                    pad.addch(char[0], char[1], char[2], char[3])
                pad.refresh(0, 0, py + 1, px + 1, y + py, x + px)
        self.stdscr.nodelay(0)

    def show(self, stuff, py=0, px=0, y=HEIGHT, x=WIDTH - 21):
        """Add a pad to the screen."""
        pad = self.newpad(y, x)
        for char in stuff:
            if char[0] < y and char[1] < x:
                if len(char) == 3:
                    pad.addch(char[0], char[1], char[2], curses.color_pair(11))
                if len(char) == 4:
                    pad.addch(char[0], char[1], char[2], char[3])
        pad.refresh(0, 0, py + 1, px + 1, y + py, x + px)

    def title_menu(self, ls, selected=0):
        """Let the user pick a built in game or go to download menu."""
        return self._internal_menu(ls, 8, selected=selected, credits=True)

    def download_menu(self, ls, selected=0):
        """Let the user pick a game from the online library."""
        return self._internal_menu(ls, 12, selected=selected)

    def gameover_menu(self, ls, selected=0):
        """Give the user some options for what to do after game ends."""
        return self._internal_menu(
            ls, 3, 6, wx=WIDTH - 30, selected=selected, controls=False,
            credits=True)

    def menu(self, ls, selected=0):
        """Give the user some options for what to do next."""
        return self._internal_menu(ls, min(8, len(ls)), selected=selected)

    def wait_for_input(self, keys=['q']):
        """Wait for the user to press a key."""
        key = ""
        key_map = {j: k for k in keys for j in self.get_ord(k)}
        while key is not None:
            key = self.stdscr.getch()
            if key in key_map:
                return key_map[key]

    def get_ord(self, key):
        """Return the key code(s) of the key."""
        if key == "ENTER":
            return [curses.KEY_ENTER, ord("\n"), ord("\r")]
        return [ord(key)]

    def _internal_menu(self, ls, y=4, py=None, selected=0, wx=WIDTH,
                       controls=True, credits=False):
        """Give the user some options, and return the selected option."""
        if py is None:
            py = HEIGHT - y - 1
        self.show_menu(ls, y, py, selected, wx, controls)
        key = ""
        while key is not None:
            key = self.stdscr.getch()
            if key in [curses.KEY_ENTER, ord("\n"), ord("\r")]:
                return selected
            if key == curses.KEY_UP:
                selected -= 1
                if selected < 0:
                    selected += len(ls)
                self.show_menu(ls, y, py, selected, wx, controls)
            if key == curses.KEY_DOWN:
                selected += 1
                if selected >= len(ls):
                    selected -= len(ls)
                self.show_menu(ls, y, py, selected, wx, controls)
            if key == ord('q'):
                if credits:
                    raise AVEQuit
                else:
                    raise AVEToMenu
            if key == ord('c') and credits:
                self.print_credits()
                self.wait_for_input(["q", "c"])
                self.print_titles()
                self.show_menu(ls, y, py, selected, wx, controls)

    def show_menu(self, ls, y, py, selected, wx, controls):
        """Display the menu."""
        if controls:
            wide = wx - 4
        else:
            wide = wx
        start = min(max(0, selected - y // 2), max(0, len(ls) - y))
        pad = self.newpad(y + 1, wx)
        for y_pos in range(y):
            if start + y_pos < len(ls):
                title = ls[start + y_pos]
            else:
                title = ""
            col = curses.color_pair(4)
            if y_pos + start == selected:
                col = curses.color_pair(5)
            pad.addstr(
                y_pos, 0,
                " " + title[: wide - 1] + " " * (wide - 1 - len(title)), col)
        if controls:
            if start > 0:
                pad.addch(0, wx - 2, "^", curses.color_pair(11))
            if start < max(0, len(ls) - y):
                pad.addch(y - 1, wx - 2, "v", curses.color_pair(11))
        pad.refresh(0, 0, py + 1, (WIDTH - wx) // 2 + 1, py + y,
                    wx + (WIDTH - wx) // 2)

    def show_titles(self, title, description, author, version):
        """Show the title screen for a game."""
        stuff = []
        y = 0
        for y in range(HEIGHT):
            if y == 0 or y == HEIGHT - 1:
                xs = range(0, WIDTH - 3, 4)
            else:
                xs = [0, WIDTH - 4]
            for x in xs:
                stuff.append((y, x, "A", curses.color_pair(6)))
                stuff.append((y, x + 1, "V", curses.color_pair(7)))
                stuff.append((y, x + 2, "E", curses.color_pair(8)))
        self.show(stuff, x=WIDTH)

        stuff = []
        pad, y, x = self.pad_with_coloured_dashes(
            title, 0, 0, HEIGHT - 2, WIDTH - 11)
        stuff += pad
        txt = "v" + str(version)
        for x, c in enumerate(txt):
            stuff.append((y - 1, WIDTH - 9 - len(txt) + x, c))
        y += 1
        txt = "Written by: " + author
        for st in range(0, len(txt), WIDTH - 9):
            for x, c in enumerate(txt[st: st + WIDTH - 9]):
                stuff.append((y, x, c))
            y += 1
        y += 1
        x = 0
        for word in description.split():
            if x + len(word) > WIDTH - 9:
                x = 0
                y += 1
                if y > HEIGHT - 5:
                    break
            for c in word:
                stuff.append((y, x, c))
                x += 1
            x += 1
        pad, y, x = self.pad_with_coloured_dashes(
            "Press <Enter> to begin. Press <q> to return to the menu.",
            HEIGHT - 4, 0, HEIGHT - 2, WIDTH - 11)
        stuff += pad

        self.show(stuff, x=WIDTH - 9, y=HEIGHT - 2, px=4, py=1)
        key = ""
        while key not in [curses.KEY_ENTER, ord("\n"), ord("\r")]:
            key = self.stdscr.getch()
            if key == ord('q'):
                raise AVEToMenu

    def pad_with_coloured_dashes(self, text, y=0, x=0, yw=HEIGHT, xw=WIDTH):
        """Pad text with coloured dashed to fill a row."""
        temp_stuff = []
        text = [" " + text[st: st + WIDTH - 11] + " "
                for st in range(0, len(text), xw)]
        cols = [curses.color_pair(6), curses.color_pair(7),
                curses.color_pair(8)]
        for t in text[: yw]:
            half = (WIDTH - 9 - len(t)) // 2
            for x in range(WIDTH - 9):
                if half <= x < half + len(t):
                    temp_stuff.append((y, x, t[x - half]))
                elif x % 4 != 3:
                    temp_stuff.append((y, x, "~", cols[x % 4]))
                else:
                    temp_stuff.append((y, x, " "))
            y += 1
        return temp_stuff, y, x



import os
from ave import config, load_game_from_file, Character
from ave.exceptions import AVEGameOver, AVEWinner
from random import choice


def test_website_move():
    numbers = {}
    inventory = []
    current_room = "test2"
    option_key = choice([0, 1, 2])
    game = load_game_from_file(os.path.join(config.root_folder,
                                            "test/games/test2.ave"))
    game.load()

    character = Character(numbers=numbers, inventory=inventory,
                          location=current_room)
    game.pick_option(option_key, character)
    try:
        text, options = game.get_room_info(character, game.currency)
    except AVEGameOver:
        return {"room": "__GAMEOVER__"}
    except AVEWinner:
        return {"room": "__WINNER__"}
    options_list = []
    for k, v in options.items():
        options_list.append((k, v))
    options_list = sorted(options_list, key=lambda x: x[0])
    inventory_text = character.get_inventory(game.items, game.currency)

    assert text == str(option_key) + " lead here"
    print(options_list)
    assert options_list[0][1] == "Continue"
    assert len(inventory_text) == 0



"""Parsing a .ave file."""

import re
from ..game import Room
from ..components import (TextWithRequirements, OptionWithRequirements,
                          NameWithRequirements)
from .string_functions import escape, unescape, clean, between
from .string_functions import escape_expression, unescape_expression
from ..components.items import Item, NumberItem
from ..components import requirements as rq
from ..components import item_giver as ig
from ..components import numbers as no


def _parse_rq(condition):
    """Parse a requirement."""
    if condition.startswith("(") and condition.endswith(")"):
        return rq.Or(*[_parse_rq(i) for i in condition[1:-1].split(",")])
    if condition.startswith("!"):
        return rq.Not(_parse_rq(condition[1:]))
    if ">" in condition or "<" in condition or "=" in condition:
        for sign in [">=", "<=", "==", "<", ">", "="]:
            if sign in condition:
                n, val = condition.split(sign, 1)
                return rq.RequiredNumber(
                    _parse_value(n), sign, _parse_value(val))
    return rq.RequiredItem(condition)


def _parse_value(value):
    """Parse a number or expression."""
    value = escape_expression(value)
    if "+" in value:
        if value[0] == "-":
            value = "0+" + value
        if "-" in value:
            value = value.replace("-", "+-")
        return no.Sum(*[_parse_value(unescape_expression(v))
                        for v in value.split("+")])
    if value.startswith("-"):
        return no.Negative(_parse_value(
            unescape_expression(value[1:])))

    if "*" in value:
        return no.Product(*[_parse_value(unescape_expression(v))
                            for v in value.split("*")])
    if "/" in value:
        return no.Division(*[_parse_value(unescape_expression(v))
                             for v in value.split("/")])

    if re.match(r"^[0-9]+$", value):
        return no.Constant(int(value))
    if re.match(r"^[0-9\.]+$", value):
        return no.Constant(float(value))

    value = unescape_expression(value)
    if value.startswith("(") and value.endswith(")"):
        return _parse_value(value[1:-1])

    if "__R__" == value:
        return no.Random()
    if value.startswith("__R__("):
        return no.Random(*[_parse_value(i)
                           for i in between(value, "(", ")").split(",")])

    return no.Variable(value)


def _parse_ig_add(item):
    """Parse an addition of an item, or addition to a number."""
    if "=" in item:
        n, value = item.split("=", 1)
        return ig.Set(n, _parse_value(value))
    if "+" in item:
        n, value = item.split("+", 1)
        return ig.Add(n, _parse_value(value))
    if "-" in item:
        n, value = item.split("-", 1)
        return ig.Remove(n, _parse_value(value))
    return ig.Add(item)


def _parse_ig_remove(item):
    """Parse the removal of an item."""
    return ig.Remove(item)


def parse_requirements(req, id_of_text="text"):
    """Parse the requirements of a line."""
    items = []
    needs = rq.Satisfied()
    pattern = re.compile(r"(\([^\)]*) ([^\)]*\))")
    while pattern.search(req) is not None:
        req = pattern.sub(r"\1,\2", req)
    pattern2 = re.compile(r" +((=|<|>)=?) +")
    while pattern2.search(req) is not None:
        req = pattern2.sub(r"\1", req)
    lsp = req.split()
    for i, j in zip(lsp[:-1:2], lsp[1::2]):
        if i == "?":
            needs = rq.And(needs, _parse_rq(j))
        elif i == "?!":
            needs = rq.And(needs, rq.Not(_parse_rq(j)))
        elif i == "+":
            items.append(_parse_ig_add(j))
        elif i == "~":
            items.append(_parse_ig_remove(j))
        else:
            raise ValueError("Unknown symbol")
    return items, needs


def parse_line(line):
    """Parse a line in a .ave file."""
    for a in ["? ", "?! ", "+ ", "~ "]:
        if line.startswith(a):
            line = " " + line
            break
    i = min([line.index(a) if a in line else len(line) for a in
             [" ? ", " ?! ", " + ", " ~ "]])
    text = unescape(line[:i])
    items, needs = parse_requirements(line[i:])
    return text, items, needs


def parse_option(line):
    """Parse a line with a destination option."""
    text, rest = line.split("=>", 1)
    text = unescape(clean(text))
    dest, req = (rest.strip(" ") + " ").split(" ", 1)
    dest = dest.strip(" ")
    items, needs = parse_requirements(req)
    if dest.startswith("__R__"):
        dests = [unescape(clean(i)) for i in
                 between(dest, "(", ")").split(",")]
        if "[" in dest:
            random = [int(i) for i in between(dest, "[", "]").split(",")]
        else:
            random = [1 for d in dests]
        return OptionWithRequirements(
            text=text, destination=dests, random=random,
            items=items, needs=needs)
    else:
        return OptionWithRequirements(
            text=text, destination=dest, items=items, needs=needs)


def parse_text_line(line):
    """Parse a line of text."""
    text, items, needs = parse_line(line)
    return TextWithRequirements(text=text, items=items, needs=needs)


def parse_name_part(line):
    """Parse the name of an item."""
    text, items, needs = parse_line(line)
    return NameWithRequirements(text=text, items=items, needs=needs)


def parse_room(id, room):
    """Parse a room."""
    room = escape(room)
    text = []
    options = []
    for line in room.split("\n"):
        line = clean(line)
        if "=>" in line:
            options.append(parse_option(line))
        elif line != "":
            text.append(parse_text_line(line))
    return Room(id=id, text=text, options=options)


def parse_item(id, item):
    """Parse an item."""
    item = escape(item)
    hidden = None
    number = False
    default = no.Constant(0)
    names = []
    for line in item.split("\n"):
        line = clean(line)
        if line.startswith("__HIDDEN__"):
            hidden = True
        elif line.startswith("__NUMBER__"):
            number = True
            if "(" in line:
                try:
                    default = _parse_value(line.split("(", 1)[1].split(")")[0])
                except ValueError:
                    default = no.Constant(0)
        elif line != "":
            if hidden is None:
                hidden = False
            names.append(parse_name_part(line))
    if number:
        return NumberItem(id=id, names=names, hidden=hidden, default=default)
    else:
        return Item(id=id, names=names, hidden=hidden)



"""Check that games are valid."""
from .error_handlers import (
    AVEFatalError, AVEError, AVEWarning, AVEInfo, AVENote)
from ..game import Character
from ..components.items import NumberItem


def check_game(game):
    """Check a game for errors."""
    errors = []
    errors += check_metadata(game)
    errors += check_first_room(game)
    errors += get_inaccessible_rooms(game)
    errors += get_undefined_rooms(game)
    errors += get_trapped_rooms(game)
    errors += get_long_options(game)
    errors += get_undefined_numbers(game)
    errors += explore_items(game)
    return errors


def check_metadata(game):
    """Check that the game has valid metadata."""
    errors = []
    if game.title == "untitled":
        errors.append(AVEError("The game's title is 'untitled' (the "
                               "default value)."))
    if game.description == "untitled":
        errors.append(AVEError("The game's description is '' (the "
                               "default value)."))
    if game.author == "anonymous":
        errors.append(AVEWarning("The game's title is 'anonymous' "
                                 "(the default value)."))
    if game.version > 1:
        errors.append(AVEInfo("The game's verion is greater than 1. "
                              "It should be an update of a preexisting "
                              "game."))
    if not isinstance(game.version, int):
        errors.append(AVEError("The game's version is not an integer."))
    if max(game.ave_version) > 0:
        errors.append(AVENote(
            "The game is set to only work on AVE"
            ">=" + ".".join(str(i) for i in game.ave_version) + "."))
    if not game.active:
        errors.append(AVEInfo("The game is deactivated."))
    return errors


def check_first_room(game):
    """Check that the first room of the game works."""
    errors = []
    if "start" not in game.rooms:
        return [AVEFatalError("There is no 'start' room.")]
    if len(game.rooms["start"].options) == 0:
        errors.append(AVEFatalError("You cannot leave the first room."))
    c = Character()
    c.reset(game.items)
    try:
        game["start"].get_text(c, game.currency)
    except:  # noqa: E722
        errors.append(AVEFatalError("Could not load text for first room."))
    try:
        game["start"].get_options(c)
    except:  # noqa: E722
        errors.append(AVEFatalError("Could not load options for first room."))
    return errors


def get_inaccessible_rooms(game):
    """Find room that cannot be reached."""
    ach = {"start"}
    for room in game.rooms.values():
        for option in room.options:
            for d in option.get_all_destinations():
                ach.add(d)
    return [AVEWarning("The room '" + i + "' is not accessible.")
            for i in game.rooms if i not in ach]


def get_long_options(game):
    """Find options whose text is too long."""
    errors = []  # 75
    for room in game.rooms.values():
        for option in room.options:
            if len(option.text) > 75:
                errors.append(AVEWarning(
                    "The option text\"" + option.text + "\" is too long."))
    return errors


def get_undefined_rooms(game):
    """Get rooms that can be reached but are not defined."""
    not_inc = set()
    for room in game.rooms.values():
        for option in room.options:
            for d in option.get_all_destinations():
                if d == "__GAMEOVER__" or d == "__WINNER__":
                    continue
                if d not in game.rooms:
                    not_inc.add(d)
    return [AVEError("The room '" + i + "' does not exist.")
            for i in not_inc]


def get_trapped_rooms(game):
    """Get rooms that there is no option to leave."""
    return [AVEError("There is no way to leave the room '" + id + "'.")
            for id, room in game.rooms.items() if len(room.options) == 0]


def get_undefined_numbers(game):
    """Get numbers that can be added to but are not defined."""
    not_def = set()
    c = Character()
    c.reset(game.items)
    for room in game.rooms.values():
        for thing in room.options + room.text:
            for item in thing.items:
                if not c.is_number(item.item) and item.value.get_value(c) != 1:
                    not_def.add(item.item)
    return [AVEWarning("The game wants to add a number to '" + i + "',"
                       " but it is not a variable.") for i in not_def]


def explore_items(game):
    """Look for items that are required but never added and other issues."""
    used = set()
    asked_for = set()
    c = Character()
    c.reset(game.items)
    for room in game.rooms.values():
        for thing in room.options + room.text:
            for item in thing.items:
                used.add(item.item)
            for item in thing.needs.get_all():
                asked_for.add(item)

    used_num = set()
    numbers = set()
    named_items = set()
    for i in game.items:
        if isinstance(i, NumberItem):
            numbers.add(i)
            if i.default.get_value(c) != 0:
                used_num.add(i)
        else:
            named_items.add(i)

    errors = []
    errors += [AVEWarning("A line in your file requires '" + i + "',"
                          " but this item is never added.")
               for i in asked_for - used - used_num]
    errors += [AVEWarning("A line in your file asks about '" + i + "',"
                          " but this number is never changed.")
               for i in (asked_for - used).intersection(numbers)]
    errors += [AVEInfo("The number '" + i + "' is defined but is not "
                       "used in the game.")
               for i in numbers - used - asked_for]
    errors += [AVEInfo("The item '" + i + "' is named but is not "
                       "used in the game.")
               for i in named_items - used - asked_for]
    errors += [AVENote("The item '" + i + "' is not named, so will be hidden "
                       "from the inventory by default.")
               for i in used.union(asked_for) - named_items - numbers]
    return errors


# Info: unnamed items will default to hidden



from math import isclose
from ave.components import numbers as no
from ave.components import item_giver as ig
from ave.components import requirements as rq
from ave import Character


def test_constant():
    c = Character()
    assert no.Constant(4).get_value(c) == 4


def test_sum():
    c = Character()
    n = no.Sum(no.Constant(1.5), no.Constant(2.5))
    assert isclose(n.get_value(c), 4)

    n = no.Sum(no.Constant(2), no.Constant(1), no.Constant(3))
    assert n.get_value(c) == 6


def test_product():
    c = Character()
    n = no.Product(no.Constant(2), no.Constant(5))
    assert n.get_value(c) == 10

    n = no.Product(no.Constant(2), no.Constant(5), no.Constant(2))
    assert n.get_value(c) == 20


def test_division():
    c = Character()
    n = no.Division(no.Constant(1), no.Constant(2))
    assert isclose(n.get_value(c), 0.5)


def text_negative():
    c = Character()
    assert no.Negative(no.Constant(2)).get_value(c) == -2


def test_variable():
    c = Character(numbers={"n": 4, "p": 5})
    assert no.Variable("n").get_value(c) == 4

    ig.Add("n", no.Constant(3)).give(c)
    assert no.Variable("n").get_value(c) == 7

    ig.Add("n", no.Variable("p")).give(c)
    assert no.Variable("n").get_value(c) == 12


def test_random():
    c = Character()

    r = no.Random()
    values = [r.get_value(c) for i in range(10)]
    assert 0 <= min(values) < max(values) <= 1

    r = no.Random(no.Constant(5))
    values = [r.get_value(c) for i in range(10)]
    assert 0 <= min(values) < max(values) <= 5

    r = no.Random(no.Constant(3), no.Constant(5))
    values = [r.get_value(c) for i in range(10)]
    assert 3 <= min(values) < max(values) <= 5


def test_items():
    c = Character()
    assert not c.has("hat")
    assert not rq.RequiredItem("hat").has(c)

    ig.Add("hat").give(c)
    assert c.has("hat")
    assert rq.RequiredItem("hat").has(c)
    assert not rq.RequiredItem("shoes").has(c)
    assert rq.Or(rq.RequiredItem("hat"),
                 rq.RequiredItem("shoes")).has(c)
    assert not rq.And(rq.RequiredItem("hat"),
                      rq.RequiredItem("shoes")).has(c)
    assert rq.Not(rq.RequiredItem("shoes")).has(c)

    ig.Add("hat").give(c)
    assert c.has("hat")

    ig.Remove("hat").give(c)
    assert not c.has("hat")

    ig.Remove("hat").give(c)
    assert not c.has("hat")


def test_numbers():
    c = Character(numbers={"n": 4})
    assert c.numbers["n"] == 4
    assert rq.RequiredNumber(no.Variable("n"), "=", no.Constant(4)).has(c)
    assert rq.RequiredNumber(no.Variable("n"), ">", no.Constant(3)).has(c)
    assert rq.RequiredNumber(no.Variable("n"), ">=", no.Constant(3)).has(c)
    assert rq.RequiredNumber(no.Variable("n"), ">=", no.Constant(4)).has(c)

    ig.Add("n").give(c)
    assert c.numbers["n"] == 5

    ig.Add("n", no.Constant(3)).give(c)
    assert c.numbers["n"] == 8

    ig.Remove("n").give(c)
    assert c.numbers["n"] == 7

    ig.Remove("n", no.Constant(2)).give(c)
    assert c.numbers["n"] == 5

    ig.Set("n", no.Constant(-1.5)).give(c)
    assert isclose(-1.5, c.numbers["n"])


def test_satisfied():
    c = Character()
    assert rq.Satisfied().has(c)
    assert not rq.Not(rq.Satisfied()).has(c)



"""The configuration of AVE."""

import os

debug = os.getenv("DEBUG")
ave_folder = os.path.dirname(os.path.realpath(__file__))

if os.path.isfile(os.path.join(ave_folder, "../VERSION")):
    # If running from folder
    root_folder = os.path.join(ave_folder, "..")
    games_folder = os.path.join(root_folder, "games")
else:
    # If running from installation
    root_folder = ave_folder
    games_folder = os.path.join(ave_folder, "_games")

screens_folder = os.path.join(ave_folder, "screens")

with open(os.path.join(root_folder, "VERSION")) as f:
    version = f.read().strip()

version_tuple = tuple(int(i) for i in version.split("."))



"""Components of a parsed AVE game."""

from .with_requirements import TextWithRequirements  # noqa: F401
from .with_requirements import NameWithRequirements  # noqa: F401
from .with_requirements import OptionWithRequirements  # noqa: F401



"""Classes for running games."""

from . import config
from .components import TextWithRequirements, OptionWithRequirements
from .components.items import NumberItem
from .exceptions import AVEGameOver, AVEWinner, AVEVersionError
from .parsing.string_functions import finalise


class Character:
    """The character playing the game."""

    def __init__(self, inventory=None, numbers=None, location=None):
        """Create the character.

        Parameters
        ----------
        inventory : list
            A list of items the character has
        numbers : dict
            The value of all numerical variables
        location : string
            The location of the character
        """
        if inventory is None:
            inventory = []
        if numbers is None:
            numbers = {}
        self.inventory = inventory
        self.numbers = numbers
        self.location = location

    def reset(self, items={}):
        """Reset the character."""
        self.inventory = []
        self.numbers = {}
        for i in items.values():
            if isinstance(i, NumberItem):
                self.numbers[i.id] = i.default.get_value(self)
        self.location = "start"

    def has(self, item):
        """Check if the character has an item."""
        if item in self.numbers:
            return self.numbers[item] > 0
        else:
            return item in self.inventory

    def add(self, item, value=1):
        """Add an item or add to a number.

        Parameters
        ----------
        item : string
            The item or variable
        value : ave.numbers.NumberItem
            The value to add
        """
        if item in self.numbers:
            self.numbers[item] += value
        elif item not in self.inventory:
            self.inventory.append(item)

    def set(self, item, value):
        """Set a number equal to a value.

        Parameters
        ----------
        item : string
            The variable
        value : ave.numbers.NumberItem
            The value to set it to
        """
        self.numbers[item] = value
        self.currency = "£"

    def remove(self, item, value=1):
        """Remove an item or take from a number.

        Parameters
        ----------
        item : string
            The item or variable
        value : ave.numbers.NumberItem
            The value to take
        """
        if item in self.numbers:
            self.numbers[item] -= value
        elif item in self.inventory:
            self.inventory.remove(item)

    def is_number(self, item):
        """Check if item is a NumberItem."""
        if not isinstance(item, str):
            raise ValueError("item must be a string")
        return item in self.numbers

    def get_inventory(self, items, currency):
        """Get the names of the character's inventory.

        Returns
        -------
        list
            The (non-hidden) items the character has
        """
        inv = []
        for i, n in self.numbers.items():
            item = items[i]
            if not item.hidden:
                name = item.get_name(self)
                if name is not None:
                    display_name = finalise(name, self.numbers, currency)
                    if item.id != "money" or currency is None:
                        display_name += ": "
                    inv.append(display_name + str(n))
        for i in self.inventory:
            if i in items:
                item = items[i]
                if not item.hidden:
                    name = item.get_name(self)
                    if name is not None:
                        inv.append(finalise(item.get_name(self), self.numbers,
                                            currency))
        return [i for i in inv if i is not None and i != ""]


class Game:
    """The Game classes that stores all the data to run the game."""

    def __init__(self, file=None, url=None,
                 filename=None, title="untitled", number=None,
                 description="", author="anonymous",
                 version=1, ave_version=(0, 0),
                 active=True):
        """Make the class.

        Parameters
        ----------
        file : string
            The full path and filename filename of the .ave file of this game
        url : string
            The url of the .ave file of this game
        filename : string
            The filename of the .ave file of this game
        title : string
            The title of the game
        number : int
            The position in the game list where this game should appear
        description : string
            A short description of the game
        author : string
            The author(s) of the game
        version : int
            The version of the game
        ave_version : tuple
            The minimum AVE version required to run this game
        active : bool
            If False, this game will only be shown in debug mode
        """
        self.id = None
        if file is not None:
            self.id = file
        if url is not None:
            self.id = url
        self.file = file
        self.filename = filename
        self.url = url
        self.number = number
        self.title = title
        self.description = description
        self.author = author
        self.active = active
        self.version = version
        self.ave_version = tuple(ave_version)
        self.currency = None
        self.rooms = None
        self.items = None

        self.options = []

    def load(self):
        """Load the full game from the file or url."""
        if config.version_tuple < self.ave_version:
            raise AVEVersionError()
        if self.file is not None:
            from .parsing.game_loader import load_full_game_from_file
            self.rooms, self.items = load_full_game_from_file(self.file)
        elif self.url is not None:
            from .parsing.game_loader import load_full_game_from_url
            self.rooms, self.items = load_full_game_from_url(self.url)
        else:
            raise ValueError("One of url and file must be set to load a game.")

        if "money" in self.items:
            c = self.items["money"].names[0].text
            if c in ["$", "£"]:
                self.currency = c

    def __getitem__(self, id):
        """Get a room with given id."""
        return self.load_room(id)

    def load_room(self, id):
        """Get a room with given id."""
        if id == "__GAMEOVER__":
            raise AVEGameOver
        if id == "__WINNER__":
            raise AVEWinner
        if id in self.rooms:
            return self.rooms[id]
        else:
            return self.fail_room()

    def pick_option(self, key, character):
        """Follow a destination option from the character's current room.

        Parameters
        ----------
        key : int
            The number of the option chosen
        character : ave.game.Character
            The character
        """
        room = self[character.location]
        o = room.options[key]
        o.get_items(character)
        character.location = o.get_destination()

    def get_room_info(self, character, currency):
        """Get the information about the character's current room.

        Parameters
        ----------
        character : ave.game.Character
            The character
        currency : str
            The currency symbol of the money variable

        Returns
        -------
        str
            The room text
        dict
            The available destination options that the character could take
        """
        room = self[character.location]
        text = room.get_text(character, currency)
        options = room.get_options(character)
        return text, {i: finalise(o.text, character.numbers, currency)
                      for i, o in options.items()}

    def fail_room(self):
        """Return a 404 error room."""
        return Room(
            "fail",
            [TextWithRequirements("You fall off the edge of the game... "
                                  "(404 Error)")],
            [OptionWithRequirements("Continue", "__GAMEOVER__")])


class Room:
    """A room in a game."""

    def __init__(self, id=None, text="", options=[]):
        """Make the room."""
        self.id = id
        self.text = text
        self.options = options

    def __str__(self):
        """Return a string."""
        return "Room with id " + self.id

    def get_text(self, character, currency):
        """Get the text for the room.

        Parameters
        ----------
        character : ave.game.Character
            The character

        Returns
        -------
        str
            The room text
        """
        lines = []
        for line in self.text:
            if line.has_requirements(character):
                line.get_items(character)
                lines.append(line.text)
        return finalise(" ".join([i for i in lines if i != ""]),
                        character.numbers, currency)

    def get_options(self, character):
        """Get the character's current destination options.

        Parameters
        ----------
        character : ave.game.Character
            The character

        Returns
        -------
        dict
            The available destination options that the character could take
        """
        return {i: o for i, o in enumerate(self.options)
                if o.has_requirements(character)}



"""Items that can be held in the character's inventory."""

from .numbers import Constant


class BaseItem:
    """A base item class."""

    def get_name(self, character):
        """Get the name of the item.

        Parameters
        ----------
        character : ave.game.Character
            The character

        Returns
        -------
        string
            The name of the item.
        """
        if self.hidden:
            return None
        out = []
        for name in self.names:
            if name.has_requirements(character):
                out.append(name.text)
        if len(out) == 0:
            return None
        return " ".join(out)


class NumberItem(BaseItem):
    """A numerical variable."""

    def __init__(self, id=id, names=[], hidden=True, default=Constant(0)):
        """Make the variable.

        Parameters
        ----------
        id : string
            The name of the variable
        names : list
            The names of the item (with requirements for possible
            alternative names)
        hidden : bool
            Should this item be hidden from the inventory list
        default : ave.numbers.Number
            The default value of the variable
        """
        self.id = id
        self.hidden = hidden
        self.names = names
        self.default = default


class Item(BaseItem):
    """An object."""

    def __init__(self, id=None, names=[], hidden=True):
        """Make the item.

        Parameters
        ----------
        id : string
            The name of the variable
        names : list
            The names of the item (with requirements for possible
            alternative names)
        hidden : bool
            Should this item be hidden from the inventory list
        """
        self.id = id
        self.hidden = hidden
        self.names = names



"""Conditions that are required to show a line, option or item name."""

from .numbers import Constant


class Requirement:
    """A base requirement."""

    def has(self, character):
        """Check if the character satisifies this."""
        raise NotImplementedError()

    def get_all(self):
        """Get all items involved in this requirement."""
        raise NotImplementedError()


class RequiredItem(Requirement):
    """The character must have an item."""

    def __init__(self, item):
        """Make the requirement."""
        self.item = item

    def has(self, character):
        """Check if the character satisifies this."""
        return character.has(self.item)

    def get_all(self):
        """Get all items involved in this requirement."""
        return [self.item]


class RequiredNumber(Requirement):
    """A numerical variable must satisfy a condition."""

    def __init__(self, v1, sign=">", v2=Constant(0)):
        """Make the requirement."""
        self.v1 = v1
        self.sign = sign
        self.v2 = v2

    def has(self, character):
        """Check if the character satisifies this."""
        v1 = self.v1.get_value(character)
        v2 = self.v2.get_value(character)
        if self.sign == ">":
            return v1 > v2
        if self.sign == "<":
            return v1 < v2
        if self.sign == ">=":
            return v1 >= v2
        if self.sign == "<=":
            return v1 <= v2
        if self.sign == "=" or self.sign == "==":
            return v1 == v2

    def get_all(self):
        """Get all items involved in this requirement."""
        return self.v1.get_all_variables() + self.v1.get_all_variables()


class Or(Requirement):
    """One of a set of Requirements must be satisfied."""

    def __init__(self, *items):
        """Make the requirement."""
        self.items = items

    def has(self, character):
        """Check if the character satisifies this."""
        for i in self.items:
            if i.has(character):
                return True
        return False

    def get_all(self):
        """Get all items involved in this requirement."""
        out = []
        for i in self.items:
            out += i.get_all()
        return out


class And(Requirement):
    """A set of Requirements must all be satisfied."""

    def __init__(self, *items):
        """Make the requirement."""
        self.items = items

    def has(self, character):
        """Check if the character satisifies this."""
        for i in self.items:
            if not i.has(character):
                return False
        return True

    def get_all(self):
        """Get all items involved in this requirement."""
        out = []
        for i in self.items:
            out += i.get_all()
        return out


class Not(Requirement):
    """The negation of another Requirement."""

    def __init__(self, item):
        """Make the requirement."""
        self.item = item

    def has(self, character):
        """Check if the character satisifies this."""
        return not self.item.has(character)

    def get_all(self):
        """Get all items involved in this requirement."""
        return self.item.get_all()


class Satisfied(Requirement):
    """This requirement is always satisfied."""

    def has(self, character):
        """Check if the character satisifies this."""
        return True

    def get_all(self):
        """Get all items involved in this requirement."""
        return []



"""Base class for displaying the game."""


class Screen:
    """The base Screen class."""

    def print_titles(self):
        """Print the AVE title page."""
        self.print_file("title")

    def print_download(self):
        """Print the game library page."""
        self.print_file("user")

    def print_credits(self):
        """Print the credits page."""
        self.print_file("credits")

    def no_internet(self):
        """Display a "You have no internet" message."""
        raise NotImplementedError()

    def clear(self):
        """Clear the screen."""
        raise NotImplementedError()

    def close(self):
        """Close the screen."""
        raise NotImplementedError()

    def put_ave_logo(self):
        """Print the AVE logo at the top of the screen."""
        raise NotImplementedError()

    def print_file(self, filename):
        """Print the contents of a text file."""

    def gameover(self):
        """Show the game over screen."""
        return self.gameend("GAME OVER")

    def winner(self):
        """Show the "you win" screen."""
        return self.gameend("YOU WIN!")

    def gameend(self, text):
        """Allow the user to play again or go back to the menu."""
        raise NotImplementedError()

    def show_inventory(self, inventory):
        """Display the inventory."""
        raise NotImplementedError()

    def type_room_text(self, text):
        """Type the room text."""
        raise NotImplementedError()

    def title_menu(self, ls, selected=0):
        """Let the user pick a built in game or go to download menu."""
        raise NotImplementedError()

    def download_menu(self, ls, selected=0):
        """Let the user pick a game from the online library."""
        raise NotImplementedError()

    def gameover_menu(self, ls, selected=0):
        """Give the user some options for what to do after game ends."""
        raise NotImplementedError()

    def menu(self, ls, selected=0):
        """Give the user some options for what to do next."""
        raise NotImplementedError()

    def wait_for_input(self, keys=['q']):
        """Wait for the user to press a key."""
        raise NotImplementedError()

    def show_titles(self, title, description, author, version):
        """Show the title screen for a game."""
        raise NotImplementedError()



import pytest
from ave.parsing import file_parsing as fp
from ave import Character


def test_requirement_parsing():
    c1 = Character()
    c2 = Character(inventory=["hat"])
    c3 = Character(inventory=["shoes"], numbers={"n": 4})

    r = fp._parse_rq("hat")
    assert not r.has(c1)
    assert r.has(c2)
    assert not r.has(c3)

    # This would be "(hat shoes)" but the parser enters the comma
    # before passing into this function.
    r = fp._parse_rq("(hat,shoes)")
    assert not r.has(c1)
    assert r.has(c2)
    assert r.has(c3)

    r = fp._parse_rq("(!hat,shoes)")
    assert r.has(c1)
    assert not r.has(c2)
    assert r.has(c3)

    r = fp._parse_rq("n")
    assert not r.has(c1)
    assert not r.has(c2)
    assert r.has(c3)

    r = fp._parse_rq("n==4")
    try:
        r.has(c1)
        assert False
    except KeyError:
        pass
    assert r.has(c3)

    r = fp._parse_rq("n>5")
    assert not r.has(c3)


@pytest.mark.parametrize(
    'txt,value', [
        ("0+1", 1),
        ("n+1", 5),
        ("n*p", 20),
        ("n*p+2*p+6", 36),
        ("n+1*p", 9),
        ("n+(1*p)", 9),
        ("n+(1)*(p)", 9),
        ("(n+1)*p", 25),
        ("(((n)+1)*p)", 25)])
def test_number_parsing(txt, value):
    c = Character(numbers={"n": 4, "p": 5})
    assert fp._parse_value(txt).get_value(c) == value


def test_random_number_parsing():
    c = Character(numbers={"n": 4, "p": 5})

    r = fp._parse_value("__R__+6")
    value = [r.get_value(c) for i in range(10)]
    assert 6 <= min(value) < max(value) <= 7

    r = fp._parse_value("__R__(4,6)+6")
    value = [r.get_value(c) for i in range(10)]
    assert 10 <= min(value) < max(value) <= 12


def test_item_giver_parsing():
    c = Character(numbers={"n": 0, "p": 1})

    assert not c.has("hat")
    fp._parse_ig_add("hat").give(c)
    assert c.has("hat")

    assert c.has("hat")
    fp._parse_ig_remove("hat").give(c)
    assert not c.has("hat")

    assert c.numbers["n"] == 0
    fp._parse_ig_add("n").give(c)
    assert c.numbers["n"] == 1
    fp._parse_ig_add("n+3").give(c)
    assert c.numbers["n"] == 4

    assert c.numbers["p"] == 1
    fp._parse_ig_add("n+p").give(c)
    assert c.numbers["p"] == 1
    assert c.numbers["n"] == 5
    fp._parse_ig_add("p-n").give(c)
    assert c.numbers["p"] == -4
    assert c.numbers["n"] == 5
    fp._parse_ig_add("p=5+6").give(c)
    assert c.numbers["p"] == 11


def test_requirement_line_parsing():
    c = Character(inventory=["hat"], numbers={"n": 2, "p": 0})

    items, needs = fp.parse_requirements("? hat + n")
    assert needs.has(c)
    items[0].give(c)
    assert c.numbers["n"] == 3

    items, needs = fp.parse_requirements("? hat ?! shoes ? (n=2 p=1)")
    assert not needs.has(c)
    items, needs = fp.parse_requirements("? hat ?! shoes ? (n=2 p=0)")
    assert needs.has(c)


def test_parse_line():
    c = Character(inventory=["hat"])

    text, items, needs = fp.parse_line("This is the text. ? hat ~ hat")
    assert text == "This is the text."
    assert needs.has(c)
    items[0].give(c)
    assert not needs.has(c)


def test_parse_option():
    c = Character()

    o = fp.parse_option("Go here => room ?! hat + shoes")
    assert o.get_destination() == "room"
    assert len(o.get_all_destinations()) == 1
    assert not c.has("shoes")
    o.get_items(c)
    assert c.has("shoes")

    o = fp.parse_option("Go here => __R__(room,garden,__GAMEOVER__)")
    assert len(o.get_all_destinations()) == 3
    for room in ["room", "garden", "__GAMEOVER__"]:
        for i in range(100):
            if o.get_destination() == room:
                break
        else:
            assert False



"""The AVE game engine."""

from .ave import AVE  # noqa: F401
from .parsing.game_loader import load_game_from_library  # noqa: F401
from .parsing.game_loader import load_game_from_file  # noqa: F401
from .game import Character, Game  # noqa: F401
from .config import version as __version__  # noqa: F401



"""Give and take items to/from a character."""

from . import numbers as no


class ItemGiver:
    """The ItemGiver class."""

    def give(self, character):
        """Give or take items."""
        raise NotImplementedError()


class Add(ItemGiver):
    """Add an item."""

    def __init__(self, item, value=no.Constant(1)):
        """Make an ItemGiver to add an item.

        Parameters
        ----------
        item : string
            An item or variable
        value : ave.numbers.Number
            The value to add to a variable
        """
        self.item = item
        self.value = value

    def give(self, character):
        """Give or take items."""
        character.add(self.item, self.value.get_value(character))


class Remove(ItemGiver):
    """Remove an item."""

    def __init__(self, item, value=no.Constant(1)):
        """Make an ItemGiver to take an item.

        Parameters
        ----------
        item : string
            An item or variable
        value : ave.numbers.Number
            The value to take from a variable
        """
        self.item = item
        self.value = value

    def give(self, character):
        """Give or take items."""
        character.remove(self.item, self.value.get_value(character))


class Set(ItemGiver):
    """Set a variable to a value."""

    def __init__(self, item, value):
        """Make an ItemGiver to set a variable to a value.

        Parameters
        ----------
        item : string
            The variable
        value : ave.numbers.Number
            The value to set it to
        """
        self.item = item
        self.value = value

    def give(self, character):
        """Give or take items."""
        character.set(self.item, self.value.get_value(character))



import pytest
import os
from ave import config, AVE, exceptions
from ave import load_game_from_file
from ave.test import check_game

config.debug = True
games_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "../games")
testgames_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "../test/games")

games = [os.path.join(games_path, filename)
         for filename in os.listdir(games_path) if filename[-4:] == ".ave"]


@pytest.mark.parametrize(
    'filename', [os.path.join(testgames_path, "hidden_test.ave")])
def test_version_checking(filename):
    try:
        game = load_game_from_file(filename)
        game.load()
        assert False
    except exceptions.AVEVersionError:
        pass

    ave = AVE()
    ave.load_games("test/games")
    for game in ave.games:
        assert game.file != filename


@pytest.mark.parametrize('filename', games)
def test_games_for_errors(filename):
    game = load_game_from_file(filename)
    game.load()

    errors = check_game(game)

    for i in range(5, 0, -1):
        ls = [e for e in errors if e.error_value == i]
        if len(ls) > 0:
            print("")
            for e in ls:
                print(e)

    # remove Info and Note errors

    errors = [i for i in errors if i.error_value > 2]

    if filename.endswith("test.ave"):
        assert len(errors) == 2
    else:
        # Assert that the worst error is a Warning or lower
        assert len(errors) == 0



import sys
import os
import setuptools

os.system("cp VERSION ave/")
os.system("rm -rf ave/_games")
os.system("cp -r games ave/_games")

if sys.version_info < (3, 4):
    print("Python 3.4 or higher required, please upgrade.")
    sys.exit(1)

with open("VERSION") as f:
    VERSION = f.read()

entry_points = {'console_scripts': ['ave = ave.__main__:run',
                                    'ave-test = ave.__main__:test_game']}

data_files = [
    ("ave/_games", [os.path.join("ave/_games", i)
                    for i in os.listdir("ave/_games")
                    if i.endswith(".ave")]),
    ("ave/screens", ["ave/screens/credits", "ave/screens/title",
                     "ave/screens/user"]),
    ("ave", ["ave/VERSION", "ave/gamelist.json"])]


if __name__ == "__main__":
    setuptools.setup(
        name="avegame",
        description="Adventure! Villainy! Excitement!",
        version=VERSION,
        author="Matthew Scroggs and Giancarlo Grasso",
        license="MIT",
        author_email="ave@mscroggs.co.uk",
        maintainer_email="ave@mscroggs.co.uk",
        url="https://github.com/AVEgame/AVE",
        packages=["ave", "ave.components", "ave.parsing", "ave.display",
                  "ave.test", "ave.screens"],
        data_files=data_files,
        include_package_data=True,
        entry_points=entry_points,
        install_requires=['windows-curses>=2.1.0 ; platform_system=="Windows"']
    )



import json
import os
from ave import config
from ave.__main__ import generate_json


def test_gamelist():
    config.debug = True
    with open(os.path.join(config.ave_folder, "gamelist.json")) as f:
        gamelist = json.load(f)
    for game in gamelist:
        game["ave_version"] = tuple(game["ave_version"])

    games = generate_json(config.games_folder)

    games.sort(key=lambda x: x["filename"])
    gamelist.sort(key=lambda x: x["filename"])
    assert gamelist == games



"""Loading a game from a .ave file."""

import re
import json
import urllib.request
from ..game import Game
from .. import config
from ..exceptions import AVENoInternet
from .string_functions import clean
from .file_parsing import parse_room, parse_item

library_json = None


def load_library_json():
    """Load the list of games in the online library."""
    global library_json
    try:
        if library_json is None:
            library_json = []
            with urllib.request.urlopen(
                    "https://www.avegame.co.uk/gamelist.json") as f:
                for game in json.load(f):
                    game["ave_version"] = tuple(game["ave_version"])
                    if game["user"]:
                        if game["ave_version"] <= config.version_tuple:
                            library_json.append(game)
    except:  # noqa: E722
        raise AVENoInternet
    return library_json


def load_full_game(text):
    """Parse the full game text."""
    rooms = {}
    for room in re.split(r"(^|\n)#", text)[1:]:
        room_id, room = re.split(r"(^|\n)%", room)[0].split("\n", 1)
        room_id = clean(room_id)
        if room_id != "":
            rooms[room_id] = parse_room(room_id, room)

    items = {}
    for item in re.split(r"(^|\n)%", text)[1:]:
        item_id, item = re.split(r"(^|\n)#", item)[0].split("\n", 1)
        item_id = clean(item_id)
        if item_id != "":
            items[item_id] = parse_item(item_id, item)

    return rooms, items


def load_game_from_file(file, filename=None):
    """Load the metadata of a game from a file."""
    title = "untitled"
    number = None
    description = ""
    author = "anonymous"
    active = True
    version = 0
    ave_version = (0, )
    with open(file) as f:
        for line in f:
            line = line.strip()
            if line.startswith("#") or line.startswith("%"):
                # Preamble has ended
                break
            if line[:2] == "==" == line[-2:]:
                title = clean(line[2:-2])
            if line[:2] == "@@" == line[-2:]:
                number = int(clean(line[2:-2]))
            if line[:2] == "--" == line[-2:]:
                description = clean(line[2:-2])
            if line[:2] == "**" == line[-2:]:
                author = clean(line[2:-2])
            if line[:2] == "vv" == line[-2:]:
                version = int(line[2:-2])
            if line[:2] == "::" == line[-2:]:
                ave_version = tuple(int(i) for i in line[2:-2].split("."))
            if line[:2] == "~~" == line[-2:]:
                if clean(line[2:-2]) == "off":
                    active = False

    return Game(file=file, filename=filename, title=title, number=number,
                description=description, author=author, active=active,
                version=version, ave_version=ave_version)


def load_game_from_library(n):
    """Load the metadata of a game from the online library."""
    info = load_library_json()[n]
    return Game(url="https://www.avegame.co.uk/download/user"
                "/" + info["filename"],
                title=info["title"], description=info["desc"],
                author=info["author"], active=info["active"],
                number=info["number"])


def load_full_game_from_file(file):
    """Load the full game from a file."""
    with open(file) as f:
        return load_full_game(f.read())


def load_full_game_from_url(url):
    """Load the full game from the online library."""
    with urllib.request.urlopen(url) as f:
        return load_full_game(f.read().decode('utf-8'))



"""Classes that make up the parsed game data."""

from random import randrange
from .requirements import Satisfied


class ThingWithRequirements:
    """A thing that needs requirements to be satisfied to be shown."""

    def __init__(self, items=[], needs=Satisfied()):
        """Make the thing."""
        self.items = items
        self.needs = needs

    def has_requirements(self, character):
        """Check is requirements are satisfied."""
        return self.needs.has(character)

    def get_items(self, character):
        """Give the character the items for and/or take items away."""
        for item in self.items:
            item.give(character)


class TextWithRequirements(ThingWithRequirements):
    """A line of test that needs requirements to be satisfied to be shown."""

    def __init__(self, text, **kwargs):
        """Make the thing."""
        self.text = text
        super().__init__(**kwargs)


class OptionWithRequirements(ThingWithRequirements):
    """An option that needs requirements to be satisfied to be shown."""

    def __init__(self, text, destination, random=None, **kwargs):
        """Make the thing."""
        self.text = text
        self.destination = destination
        if random is None:
            self.random = False
        else:
            self.random = True
            self.probabilities = random
        super().__init__(**kwargs)

    def get_all_destinations(self):
        """Get a list of all destinations that this options could lead to.

        Randomness will lead to multiple destinations from a single option.
        """
        if self.random:
            return self.destination
        else:
            return [self.destination]

    def get_destination(self):
        """Get the destination when this option is chosen."""
        if self.random:
            n = randrange(sum(self.probabilities))
            total = 0
            for d, i in zip(self.destination, self.probabilities):
                total += i
                if total > n:
                    return d
        else:
            return self.destination


class NameWithRequirements(ThingWithRequirements):
    """The name of an item that needs requirements to be satisfied."""

    def __init__(self, text, **kwargs):
        """Make the thing."""
        self.text = text
        super().__init__(**kwargs)
